import asyncio
import asyncssh
import sys
from pathlib import Path
import re
import logging
import subprocess


async def run_scp_copy_dask_prefs(filename: Path, server: str, key: Path, remote_folder: Path):
    print("remote_path: ", remote_folder)
    await asyncssh.scp(str(filename), (server, remote_folder),
                       client_keys=key)


async def run_launch_dask_head_node(gateway: str, key: Path, prefs: dict, saver, home_folder: Path):
    command1 = f"conda activate {prefs['remote_conf']['env']}"
    command2 = (f"sbatch -c {prefs['remote_conf']['cores']} -p {prefs['remote_conf']['queue']} "
                f"{prefs['remote_conf']['command']} {prefs['remote_conf']['env']} {home_folder}")
    command = command1 + ' ; ' + command2
    print(f"command: {command}")
    async with asyncssh.connect(gateway, client_keys=key) as conn:
        try:
            result = await conn.run(command, check=True)
            print(result)
            job_id_match = re.match(r'Submitted batch job (?P<JOB_ID>\d+)', result.stdout)
            if job_id_match:
                head_node_id = job_id_match.groupdict()["JOB_ID"]
                saver.write_prefect_job_id(head_node_id)
            else:
                logging.warning(f"Could not parse job_id:")
                logging.warning(f"\t{result.stdout}")
        except asyncssh.ProcessError as e:
            print(e)


async def run_watch_and_report_ports(gateway: str, key: Path, rel_folder):
    pattern = re.compile(r"tcp://(?P<host>[\d\.]+):(?P<DASK_PORT>\d+)\s+(?P<DASHBOARD_PORT>\d+)$",
                               re.MULTILINE)
    rel_pidfile = rel_folder / "pidfile"
    rel_stdout = rel_folder / "stdout"
    command1 = f"while [ ! -s {rel_pidfile} ] ; do sleep 1; done \n"
    command2 = f"while [ ! -s {rel_stdout} ]; do sleep 1; done \n"
    command3 = f"cat {rel_stdout} \n"
    result = None
    async with asyncssh.connect(gateway, client_keys=key) as conn:
        await conn.run(command1, check=True)
        await conn.run(command2, check=True)
        result = await conn.run(command3, check=True)
        my_match = pattern.match(result.stdout)
        if my_match:
            result = my_match.groupdict()
        else:
            logging.warning(f"Regular expression couldn't parse ports from :")
            logging.warning(f"\t{result.stdout}")
    return result


async def shutdown_dask_head_node(gateway: str, key: Path, job_id: int, pidfile: Path):
    command = f"scancel {job_id}; rm {pidfile}\n"
    async with asyncssh.connect(gateway, client_keys=key) as conn:
        await conn.run(command, check=True)

# async def query_jobs_running(gateway: str, key: Path, user: str, keyword: str):
#     command = f"squeue -u {user}"
#     async with asyncssh.connect(gateway, client_keys=key) as conn:
#         result = await conn.run(command, check=True)
#     mtch = r"".match()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import importlib
import logging
from typing import Dict, List, Union

###############################################################################

log = logging.getLogger(__name__)

###############################################################################


def load_custom_object(module_path: Union[str, List[str]], object_name: str, object_kwargs: Dict) -> object:
    """
    Load a custom object with kwargs.
    Parameters
    ----------
    module_path: Union[str, List[str]]
        Python module path or list of path parts to a custom module. Ex: "cptools.pipeline"
    object_name: str
        Name of the object to retrieve from the module. Ex: "Pipeline"
    object_kwargs: Dict
        Any kwargs to pass to the object.
    Returns
    -------
    obj: object
        The initialized object.
    """
    # Convert module path to string
    if isinstance(module_path, list):
        module_path = ".".join(module_path)

    # Load target module
    mod = importlib.import_module(module_path)
    obj = getattr(mod, object_name)
    obj = obj(**object_kwargs)

    # Log
    log.debug(f"Using object: {type(obj)}")

    return obj



import pytest
from scheduler_tools.PrefectPreferences import PrefectPreferences


def test_user_pref():
    pp = PrefectPreferences()


def test_given(data_dir):
    pp = PrefectPreferences(path=data_dir/"ssh.json")


def test_url(data_dir):
    pp = PrefectPreferences(path=data_dir/"ssh.json")
    assert pp.gateway_url == "slurm-master.corp.alleninstitute.org"


def test_username(data_dir):
    pp = PrefectPreferences(path=data_dir / "ssh.json")
    assert pp.username == "jamies"


def test_identity_file(data_dir):
    pp = PrefectPreferences(path=data_dir/"ssh.json")
    assert pp.identity_file




import getpass
from typing import Union


def default_dask_prefs(remote_user: Union[str, type(None)] = None):
    # in no remote user is provided assume the remote username is the same as the local one
    if remote_user is None:
        remote_user = getpass.getuser()

    # setup the dictionary
    dask_prefs = {'cluster_obj_name': {},
                  'host_conf': {},
                  'worker_conf': {},
                  'adapt_conf': {},
                  'remote_conf': {}}

    # specify the command to specify the type of cluster to launch and the module in which it lives
    dask_prefs['cluster_obj_name'] = {'module': 'dask_jobqueue', 'object': 'SLURMCluster'}

    # dictionary of setting to pass to sbatch (to define host resources)
    dask_prefs['host_conf'] = {'queue': 'aics_cpu_general', 'cores': 2, 'memory': '8GB',
                                           'walltime': '02:00:00'}

    # These are currently redundant to host prefs but this might change with other launchers
    # specify the arguments to the SLURMCluster specified above
    dask_prefs['worker_conf'] = {'queue': 'aics_cpu_general', 'cores': 2, 'memory': '8GB',
                                           'walltime': '02:00:00'}

    # specify the worker parameters here
    dask_prefs['adapt_conf'] = {'minimum_jobs': 2, 'maximum_jobs': 40}

    # spicify the remote conda environment,
    # the remote command to fire (this shouldn't be changed),
    # the remote settings folder
    dask_prefs['remote_conf']['env'] = 'dask-scheduler'
    dask_prefs['remote_conf']['command'] = 'setup_and_spawn.bash'
    dask_prefs['remote_conf']['path'] = f"/home/{remote_user}/.aics_dask"
    dask_prefs['remote_conf']['queue'] = 'aics_cpu_general'
    dask_prefs['remote_conf']['cores'] = 2

    return dask_prefs



from pathlib import Path
from scheduler_tools.types import PrefDict


class PrefectPreferences:
    """
    This class handles reading of a ~/.prefect/ssh.json file. This file has settings for the
    name of the gateway, the username to authenticate with, the path to the local ssh identity
    file.
    """

    def __init__(self, prefs: PrefDict):
        """

        :param prefs:
        """
        print(prefs)
        p_localfolder = Path(prefs['localfolder'])
        print("1: ")
        if not p_localfolder.exists():
            p_localfolder.mkdir(parents=True)

        print("2: ")
        self._path = p_localfolder

        print("3: ")
        self._data = prefs

    def default_path(self) -> Path:
        return self._path

    @property
    def gateway_url(self):
        return self._data['gateway']['url']

    @property
    def username(self):
        return self._data['gateway']['user']

    @property
    def identity_file(self):
        return self._data['gateway']['identityfile']

    @property
    def known_hosts(self):
        return Path('~/.ssh/known_hosts').expanduser()

    def write_ssh_pid(self, pid):
        with open(str(self.ssh_pid_path()), 'w') as fp:
            fp.write(str(pid))

    def read_ssh_pid(self) -> [str, type(None)]:
        pid = None
        if self.ssh_pid_path().expanduser().exists():
            pid = open(str(self.ssh_pid_path().expanduser()), 'r').read()
        return pid

    def remove_ssh_pid(self):
        if self.ssh_pid_path().exists():
            self.ssh_pid_path().unlink()

    def ssh_pid_path(self):
        return self.default_path().expanduser() / "ssh_pid.txt"

    def cluster_job_id_path(self):
        return self.default_path().expanduser() / "cluster_job_id.txt"

    def read_prefect_job_id(self) -> [str, type(None)]:
        job_id = None
        if self.cluster_job_id_path().exists():
            job_id = open(str(self.cluster_job_id_path().expanduser()), 'r').read()
        return job_id

    def write_prefect_job_id(self, job_id):
        print(f"jobid: {job_id}")
        with open(str(self.cluster_job_id_path().expanduser()), 'w') as fp:
            fp.write(str(job_id))

    def remove_prefect_job_id(self):
        if self.cluster_job_id_path().exists():
            self.cluster_job_id_path().unlink()

    def cluster_pid_path(self):
        # this needs to be made dynamic
        return self.default_path().relative_to(Path().home()) / "pidfile"

    @property
    def local_dask_port(self):
        return self._data['dask_port']

    @property
    def local_dashboard_port(self):
        return self._data['dashboard_port']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Configuration for tests! There are a whole list of hooks you can define in this file to run before, after, or to mutate
how tests run. Commonly for most of our work, we use this file to define top level fixtures that may be needed for tests
throughout multiple test files.

In this case, while we aren't using this fixture in our tests, the prime use case for
something like this would be when we want to preload a file to be used in multiple tests. File reading can take time,
so instead of re-reading the file for each test, read the file once then use the loaded content.

Docs: https://docs.pytest.org/en/latest/example/simple.html
      https://docs.pytest.org/en/latest/plugins.html#requiring-loading-plugins-in-a-test-module-or-conftest-file
"""

import json
from pathlib import Path
from typing import Dict

import pytest


@pytest.fixture
def data_dir() -> Path:
    return Path(__file__).parent / "data"

# Fixtures can simply be added as a parameter to the other test or fixture functions to expose them.
# If we had multiple tests that wanted to use the contents of this file, we could simply add "loaded_example_values" as
# a parameter for each test.
@pytest.fixture
def loaded_example_values(data_dir) -> Dict[str, int]:
    with open(data_dir / "example_values.json", "r") as read_in:
        return json.load(read_in)



import asyncio
import json
import logging
import os
import re
import subprocess
import sys
from pathlib import Path

import asyncssh
import psutil

from scheduler_tools.PrefectPreferences import PrefectPreferences
from scheduler_tools.ScpDaskPrefs import run_scp_copy_dask_prefs, \
    run_launch_dask_head_node, run_watch_and_report_ports, shutdown_dask_head_node
from scheduler_tools.types import PathDict, PrefDict
from scheduler_tools.default_dask_prefs import default_dask_prefs
from scheduler_tools.default_ssh_prefs import default_ssh_prefs


class Connector:

    def __init__(self, dask_prefs: dict = None, prefs: PathDict = None):

        self._dask_scheduler_ip = None
        self._dask_scheduler_port = None
        self._dask_dashboard_port = None
        self._tunnel = None
        self._pattern = re.compile(r"tcp://(?P<host>[\d\.]+):(?P<DASK_PORT>\d+)\s+(?P<DASHBOARD_PORT>\d+)$",
                                   re.MULTILINE)

        if prefs is None:
            prefs = default_ssh_prefs()
        elif isinstance(prefs, (Path, str)):
            localfolder = Path(prefs).parent
            if Path(prefs).expanduser().exists():
                with open(Path(prefs).expanduser(), 'r') as fp:
                    prefs = json.load(fp)
                print(f"localfolder: {localfolder}")
                if 'localfolder' not in prefs.keys():
                    prefs['localfolder'] = localfolder
            else:
                raise FileNotFoundError(f"{prefs} file does not exists.")
            print("survived elsif")
        print("in pref")
        self._pref = PrefectPreferences(prefs)
        print("from pref")

        if dask_prefs is None:
            self._dask_prefs = default_dask_prefs()
        else:
            self._dask_prefs = dask_prefs
        print("survived init")
        self.serialize_and_push_prefs()

    def serialize_and_push_prefs(self):
        # serialize the json prefs
        dask_prefs_path = self._pref.default_path().expanduser() / "dask_prefs.json"
        print(f"in serialize: {dask_prefs_path}")
        with open(str(dask_prefs_path), 'w') as fp:
            json.dump(self._dask_prefs, fp)
        # scp json to cluster
        try:
            print(f"relp: {self._dask_prefs['remote_conf']['path']}")
            relative_path = Path(self._dask_prefs['remote_conf']['path'])
            print(f"rel path: {relative_path}")
            asyncio.get_event_loop().run_until_complete(run_scp_copy_dask_prefs(dask_prefs_path,
                                                                                self._pref.gateway_url,
                                                                                self._pref.identity_file,
                                                                                relative_path))
        except (OSError, asyncssh.Error) as exc:
            sys.exit('SCP connection failed: ' + str(exc))
        print("survived serialize")

    def run_command(self):
        self.serialize_and_push_prefs()
        remote_home = self._dask_prefs['remote_conf']['path']
        try:
            asyncio.get_event_loop().run_until_complete(run_launch_dask_head_node(self._pref.gateway_url,
                                                                                  self._pref.identity_file,
                                                                                  self._dask_prefs,
                                                                                  self._pref,
                                                                                  remote_home))
        except (OSError, asyncssh.Error) as exc:
            sys.exit('Launch dask cluster failed: ' + str(exc))

    def stop_forward_if_running(self, query=True):
        pid_str = self._pref.read_ssh_pid()
        # if there is no pid or the pid isn't valid return
        if pid_str is None or not psutil.pid_exists(int(pid_str)):
            self._pref.remove_ssh_pid()
            return
        # if the pid doesn't match the command issued
        ppid = psutil.Process(pid=int(pid_str))
        children = ppid.children(recursive=True)
        if ppid and children:
            logging.info(f"Shutting down PID({ppid.pid}): {ppid.name()}")
            for child in children:
                logging.info(f"Shutting down PID({child.pid}): {child.name()}")
            remove = True
            if query:
                data = input("Shutdown the above processes? [YES/no]")
                pat = r'\s*(no|NO|No)\s*'
                remove = not re.match(pat, data)
            if remove:
                for child in children:
                    child.kill()
                ppid.kill()
        # remove the file
        self._pref.ssh_pid_path().unlink()

    @property
    def local_dask_port(self):
        return self._pref.local_dask_port
    
    @property
    def local_dashboard_port(self):
        return self._pref.local_dashboard_port

    def forward_ports(self):
        rel_folder = self._pref.default_path().expanduser().relative_to(Path().home())
        info = asyncio.get_event_loop().run_until_complete(
            run_watch_and_report_ports(self._pref.gateway_url, self._pref.identity_file, rel_folder)
        )

        ssh_command = (f"ssh-add {self._pref.identity_file} ; "
                       f"ssh -o StrictHostKeyChecking=no "
                       f"-N -A -J {self._pref.gateway_url} "
                       f"-L {self._pref.local_dask_port}:{info['host']}:{info['DASK_PORT']} "
                       f"-L {self._pref.local_dashboard_port}:{info['host']}:{info['DASHBOARD_PORT']} "
                       f"{self._pref.username}@{info['host']}"
                       )

        print(ssh_command)
        self._tunnel = subprocess.Popen(ssh_command, shell=True, executable='/bin/bash')
        os.environ['DASK_PORT'] = info['DASK_PORT']
        print(f"\nTo connect to remote server use \n\tDask:\t \t localhost:{self._pref.local_dask_port}\n"
              f"\tDashboard:\t localhost:{self._pref.local_dashboard_port}")
        self._pref.write_ssh_pid(pid=self._tunnel.pid)

    def stop_dask(self):
        job_id = self._pref.read_prefect_job_id()
        pidfile = Path(self._dask_prefs['remote_conf']['path']) / "pidfile"
        if job_id is None:
            logging.warning("Job ID is None")
            return
        logging.info(f"Cancelling Slurm Head Node: {job_id}")
        try:
            asyncio.get_event_loop().run_until_complete(shutdown_dask_head_node(self._pref.gateway_url,
                                                                                self._pref.identity_file,
                                                                                job_id,
                                                                                pidfile))
        except asyncssh.process.ProcessError as e:
            logging.warn(f"it looks like the remote process was already shutdown: {e}")

        self.stop_forward_if_running(query=False)

    def __enter__(self):
        self.run_command()
        self.stop_forward_if_running()
        self.forward_ports()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop_dask()
        self.stop_forward_if_running()



from .types import PrefDict
from pathlib import Path
from os import uname


def default_ssh_prefs():
    ssh_prefs = PrefDict()
    ssh_prefs['localfolder'] = Path('~/.aics_dask').expanduser()
    ssh_prefs['gateway'] = {'url': 'slurm-master', 'user': uname(),
                            'identityfile': ssh_prefs['localfolder']/'.ssh/id_rsa'}
    ssh_prefs['dask_port'] = 34000
    ssh_prefs['dashboard_port'] = 8787
    return ssh_prefs



from datetime import datetime
from pathlib import Path
from time import sleep

from aicsdaemon import Daemon

from .load_custom_obj import load_custom_object
from .types import Pathlike


class AicsPrefectDaemon(Daemon):

    def __init__(self, slurm_prefs: dict, pidfile: Pathlike, stdin: Pathlike = None, stdout: Pathlike = None,
                 stderr: Pathlike = None, foreground: bool = False):
        super(AicsPrefectDaemon, self).__init__(pidfile=pidfile,
                                                stdout=stdout,
                                                stdin=stdin,
                                                stderr=stderr,
                                                foreground=foreground)
        print("AicsPrefectDaemon: pidfile: ", pidfile)
        print(f"prefs {slurm_prefs}")
        open(stdout, 'w').close() if stdout else None

        localdir = Path(pidfile).parent / "logs" / datetime.now().strftime("%Y%m%d_%H:%M:%S")

        self._cluster_obj_name = slurm_prefs['cluster_obj_name']

        self._cluster_prefs = slurm_prefs['host_conf'].copy()
        self._cluster_prefs['local_directory'] = localdir
        self._cluster_prefs['log_directory'] = localdir

        self._adapt_prefs = slurm_prefs['adapt_conf'].copy()

    def run(self):

        object_kwargs = self._cluster_prefs
        if self._adapt_prefs is None:
            object_kwargs = {}

        cluster = load_custom_object(module_path=self._cluster_obj_name['module'],
                                     object_name=self._cluster_obj_name['object'],
                                     object_kwargs=object_kwargs)

        if self._adapt_prefs is not None:
            cluster.adapt(**self._adapt_prefs)

        print(f"{cluster.scheduler_info['address']}")
        print(f"{cluster.scheduler_info['services']['dashboard']}", flush=True)

        # this enables us to put something in stdin and if so this loop will
        # exit. Thus using the stdin="filepath" argument gives us a clean way to
        # shutdown the process
        mtime = self.stdin.stat().st_mtime
        while mtime == self.stdin.stat().st_mtime:
            sleep(5)





from typing import Union
from typing_extensions import TypedDict
from pathlib import Path


class PrefDict(TypedDict):
    GateWay = TypedDict('gateway', {'url': str, 'user': str, 'identityfile': Union[str, Path]})
    gateway: GateWay
    localfolder: str
    dask_port: int
    dashboard_port: int


Pathlike = Union[Path, str]
PathDict = Union[Pathlike, PrefDict, type(None)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This sample script will get deployed in the bin directory of the
users' virtualenv when the parent module is installed using pip.
"""

import argparse
import logging
import sys
import traceback
from pathlib import Path

from scheduler_tools.aics_prefect_daemon import AicsPrefectDaemon

###############################################################################

log = logging.getLogger()
logging.basicConfig(level=logging.INFO,
                    format='[%(levelname)4s:%(lineno)4s %(asctime)s] %(message)s')

###############################################################################

###############################################################################


class Args(argparse.Namespace):

    DEFAULT_FIRST = 10
    DEFAULT_SECOND = 20

    def __init__(self):
        # Arguments that could be passed in through the command line
        self.first = self.DEFAULT_FIRST
        self.second = self.DEFAULT_SECOND
        self.debug = False
        #
        self.__parse()

    def __parse(self):
        pth = Path("~/.prefect").expanduser().resolve()
        p = argparse.ArgumentParser(prog='stop_adaptable_dask_cluster',
                                    description='shutdown the adaptable dask cluster daemon running for prefect')
        p.add_argument('-p', '--pidfile', help='where to lookup the pid', dest='pidfile', type=Path,
                       default=pth / "pidfile")
        p.add_argument('--debug', action='store_true', dest='debug', help=argparse.SUPPRESS)
        p.parse_args(namespace=self)


###############################################################################

def main():
    dbg = True
    try:
        args = Args()
        dbg = args.debug

        print("pid file => ", args.pidfile)
        daemon = AicsPrefectDaemon(pidfile=args.pidfile)
        daemon.stop()

    except Exception as e:
        log.error("=============================================")
        if dbg:
            log.error("\n\n" + traceback.format_exc())
            log.error("=============================================")
        log.error("\n\n" + str(e) + "\n")
        log.error("=============================================")
        sys.exit(1)


###############################################################################
# Allow caller to directly run this module (usually in development scenarios)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import find_packages, setup

with open("README.md") as readme_file:
    readme = readme_file.read()

test_requirements = [
    "codecov",
    "flake8",
    "pytest",
    "pytest-cov",
    "pytest-raises",
]

setup_requirements = [
    "pytest-runner",
]

dev_requirements = [
    "bumpversion>=0.5.3",
    "coverage>=5.0a4",
    "flake8>=3.7.7",
    "ipython>=7.5.0",
    "m2r>=0.2.1",
    "pytest>=4.3.0",
    "pytest-cov==2.6.1",
    "pytest-raises>=0.10",
    "pytest-runner>=4.4",
    "Sphinx>=2.0.0b1",
    "sphinx_rtd_theme>=0.1.5",
    "tox>=3.5.2",
    "twine>=1.13.0",
    "wheel>=0.33.1",
]

interactive_requirements = [
    "altair",
    "jupyterlab",
    "matplotlib",
]

requirements = [
    "asyncssh",
    "asyncio",
    "aicsdaemon",
    "bokeh",
    "dask",
    "dask_cuda",
    "dask_jobqueue",
    "typing_extensions"
]

extra_requirements = {
    "test": test_requirements,
    "setup": setup_requirements,
    "dev": dev_requirements,
    "interactive": interactive_requirements,
    "all": [
        *requirements,
        *test_requirements,
        *setup_requirements,
        *dev_requirements,
        *interactive_requirements
    ]
}

setup(
    author="Gregory Johnson",
    author_email="gregj@alleninstitute.org",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: Free for non-commercial use",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description="Tools for composing interdependent tasks on a scheduler (i.e. SLURM)",
    entry_points={
        "console_scripts": [
            "spawn_dask_cluster=scheduler_tools.bin.spawn_dask_cluster:main",
            "shutdown_dask_cluster=scheduler_tools.bin.shutdown_dask_cluster:main",
            "spawn_adaptable_dask_cluster=scheduler_tools.bin.spawn_adaptable_dask_cluster:main",
            "stop_adaptable_dask_cluster=scheduler_tools.bin.stop_adaptable_dask_cluster:main",
        ],
    },
    install_requires=requirements,
    license="Allen Institute Software License",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="scheduler_tools",
    name="scheduler_tools",
    packages=find_packages(),
    python_requires=">=3.6",
    scripts=[
        'scheduler_tools/bin/setup_and_spawn.bash'
    ],
    setup_requires=setup_requirements,
    test_suite="scheduler_tools/tests",
    tests_require=test_requirements,
    extras_require=extra_requirements,
    url="https://github.com/AllenCellModeling/scheduler_tools",
    # Do not edit this string manually, always use bumpversion
    # Details in CONTRIBUTING.rst
    version="0.1.5",
    zip_safe=False,
)



import subprocess
import os
import re

# QSUB to SLURM Mappings
# queue_name    -> partition
# walltime      -> time
SBATCH_TEMPLATE = """#!/bin/bash

#SBATCH --partition={partition}
#SBATCH --time={walltime}
#SBATCH --output={out_err_format}.out
#SBATCH --error={out_err_format}.err
{conditional_array_string}
{conditional_deps_string}

srun {task_name} {conditional_verbose}
"""


def submit_jobs(files, json_obj, tmp_file_name="tmp.script"):

    # Cast files to a list
    if not isinstance(files, list):
        files = [files]

    # Determine array structure
    task_path_sep = files[0].split("/")
    shell_script_name = task_path_sep[-1]

    # Only manipulate the task name if it is an array operation
    # This regex will find scripts that match similar patterns such as: `im2feats_123.sh`
    # The important part here is that we are looking for `_***.sh`, if the script name has this
    # it indicates that the script is part of an job array operation.
    if re.match(r"^[a-zA-Z0-9]*_[0-9]*\.sh$", shell_script_name):
        # Great, this is an array job split the task id from the job type
        task_parts = shell_script_name.split("_")
        # Get the parent path
        original_path = "/".join(task_path_sep[:-1])

        # Use the job type to create formatted strings to be attached to sbatch
        # This will result in something like the following for task_name:
        # /path/to/original/parent/im2feats_$SLURM_ARRAY_TASK_ID.sh
        # and something like the following for out_err_format:
        # /path/to/original/parent/im2feats_%a.sh
        task_name = f"{original_path}/{task_parts[0]}_$SLURM_ARRAY_TASK_ID.sh"
        out_err_format = f"{original_path}/{task_parts[0]}_%a.sh"
    else:
        # This isn't part of a job array operation
        # The task name and out_err_format will then just be the filename
        task_name = files[0]
        out_err_format = files[0]

    # Retain max tasks for later
    max_tasks = None
    if len(files) > 1:
        # Split the file path, get the last component, split the name, get the suffix, split the suffix, get number
        # For a collection of files that look like the following:
        # ["/allen/aics/modeling/gregj/results/ipp/scp_2019_03_25/scripts/im2feats_0.sh",
        #  "/allen/aics/modeling/gregj/results/ipp/scp_2019_03_25/scripts/im2feats_1.sh"
        #  ...
        #  "/allen/aics/modeling/gregj/results/ipp/scp_2019_03_25/scripts/im2feats_746.sh"]
        # This will find the max integer value of these files. In the above example it would result in 746.
        # This is done by split the file by the os filepath sep,
        # retrieving the filename (the last item in the created list),
        # splitting the filename into its task type and task array value using "_"
        # retrieving the value by selecting the last item in the task details array
        # and finally removing the suffix by splitting my "." and selecting the first item
        max_tasks = max([int(file.split("/")[-1].split("_")[-1].split(".")[0]) for file in files])
        array_string = f"#SBATCH --array=0-{max_tasks}%{json_obj['max_parallel_jobs']}"
    else:
        array_string = ""

    # Create deps string
    if len(json_obj.get("deps", [])) > 0:
        # As dependecies are created and added to the json object as run_scp_jobs progresses,
        # this checks for the current value of that dependency list and if there is one, as in the actual depencency
        # key exists and has at least one item, it will generate an sbatch dependency string to be attached to the
        # sbatch headers.
        # sbatch dependencys follow the template "<JOB_ID>:<JOB_ID>:<JOB_ID>...:<JOB_ID>"
        deps_string = "#SBATCH --dependency=afterany:{}".format(
            ":".join([str(dep_job) for dep_job in json_obj["deps"]])
        )
    else:
        deps_string = ""

    # Create verbose conditional
    if json_obj.get("debug", "false").lower() == "true":
        debug = "--verbose"
    else:
        debug = ""

    sbatch_str = SBATCH_TEMPLATE.format(
        partition=json_obj["partition"],
        walltime=json_obj["walltime"],
        conditional_array_string=array_string,
        conditional_deps_string=deps_string,
        task_name=task_name,
        out_err_format=out_err_format,
        conditional_verbose=debug
    )

    tmp_file = json_obj["script_dir"] + os.sep + tmp_file_name
    with open(tmp_file, "w") as write_out:
        write_out.write(sbatch_str)

    # Push sbatch file to slurm queue
    proc = subprocess.Popen("sbatch " + tmp_file, stdout=subprocess.PIPE, shell=True)
    out, err = proc.communicate()

    # As slurm accepts jobs, it returns the job id in a string along with some other text
    # This will find the job id returned by spliting the string by space and getting the last one for each line ("\n")
    # of returned text from the slurm sbatch request.
    # However, as this is somewhat legacy code where previously multiple jobs were submitted by now in SLURM a single
    # job is submitted that spawns a lot of additional jobs due to array tasking, we really only care about the first
    # one (or really only a single job id will ever be returned).
    # Regardless, due to this change we need to construct all the actual job_ids that are created from the sbatch and
    # not just the original job that spawned all the sub jobs.
    # To do this we use the max_tasks value we stored earlier and construct a range over it.
    job_id = [int(token.split(" ")[-1]) for token in out.decode("utf8").split('\n') if token != ""][0]
    if max_tasks:
        job_ids = [f"{job_id}_{i}" for i in range(max_tasks + 1)]
    else:
        job_ids = [job_id]

    return job_ids



# -*- coding: utf-8 -*-

"""Top-level package for scheduler_tools."""

__author__ = "Gregory Johnson"
__email__ = "gregj@alleninstitute.org"
# Do not edit this string manually, always use bumpversion
# Details in CONTRIBUTING.md
__version__ = "0.1.5"


def get_module_version():
    return __version__


from .example import Example  # noqa: F401
from .connector import Connector  # noqa: F401




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Imports should be grouped into:
# Standard library imports
# Related third party imports
# Local application / relative imports
# in that order

# Standard library
import logging
from typing import Any, Tuple

# Third party

# Relative

###############################################################################

log = logging.getLogger(__name__)

###############################################################################


class Example(object):
    """
    This is an example object. Use this example for ideas on how to write doc strings, use logging, construct
    objects. This is not an exhaustive example but a decent start.

    Parameters
    ----------
    init_value: int
        An integer value to initialize the object with.
    """

    # Static methods are available to the user regardless of if they have initialized an instance of the class
    # They are useful when you have small portions of code that while relevant to the class may not depend on
    # entire class state.
    # In this case, this function isn't incredibly valuable outside of the usage of this class and therefore
    # we use the "Python" standard of prefixing the method with an underscore
    @staticmethod
    def _check_value(val: Any):
        """
        Check that the value is an integer. If not, raises a ValueError.
        """
        if not isinstance(val, int):
            raise ValueError(f"Provided value: {val} (type: {type(val)}, is not an integer.)")

    def __init__(self, init_value: int = 10):
        # Check initial value
        self._check_value(init_value)

        # Set values
        self.current_value = init_value
        self.old_value = None

    def update_value(self, new_value: int) -> int:
        """
        Save old value and set new value.

        Parameters
        ----------
        new_value: int
            The new value to assign to the object.

        Returns
        -------
        old_value: int
            The previous value stored in the object.
        """
        # Check new value before assign
        self._check_value(new_value)

        # Passed, now assign
        self.old_value = self.current_value
        self.current_value = new_value
        log.info(f"Updating value from {self.old_value} to {self.current_value}")
        return self.old_value

    def get_value(self) -> int:
        """
        Get the current value.

        Returns
        -------
        current_value: int
            The current value stored in the object.
        """
        return self.current_value

    def get_previous_value(self) -> int:
        """
        Get the previous value.

        Returns
        -------
        previous_value: int
            The previous value stored in the object.
        """
        return self.old_value

    # And example of a property accessor
    # When using this object, this "function" can be called using property / attribute style
    # These are useful when you want to hide that computation or IO is happening from the user.
    # Usually, these are used when you need to lazy load something or have "immutability" of an object property.
    # ```
    # e = Example(10)
    # stored = e.values
    # ```
    @property
    def values(self) -> Tuple[int]:
        """
        Get both values stored in the object as a tuple of integers.

        Returns
        -------
        values: Tuple[int]
            The current and old values stored in a tuple in (current_value, old_value) order.
        """
        return (self.current_value, self.old_value)

    def __str__(self):
        return f"<Example [current: {self.current_value}, previous: {self.old_value}]>"

    # Representation's (reprs) are useful when using interactive Python sessions or when you log an object
    # They are the shorthand of the object state. In this case, our string method provides a good representation.
    def __repr__(self):
        return str(self)
